#include "FileSelector.hpp"

FileSelector::FileSelector(const QString &goText, PassedOptions *opts, QWidget *parent)
: QWidget(parent) {
	QHBoxLayout *outerLayout = new QHBoxLayout;
	QVBoxLayout *innerLayout = new QVBoxLayout;

	addButton = new QPushButton(tr("Add &files"));
	addDirButton = new QPushButton(tr("Add &directory"));
	removeButton = new QPushButton(tr("&Remove"));
	goButton = new QPushButton(goText);
	options = opts;

	theChosenOnes = new QListView;
	theChosenOnes->setSelectionMode(QAbstractItemView::ExtendedSelection);
	theChosenOnes->setEditTriggers(QAbstractItemView::NoEditTriggers);
	theChosenOnes->setSelectionRectVisible(true);
	theChosenModel = new QStringListModel;
	theChosenOnes->setModel(theChosenModel);
	
	outerLayout->addWidget(theChosenOnes,1);

	innerLayout->addWidget(addButton);
	innerLayout->addWidget(addDirButton);
	innerLayout->addWidget(removeButton);
	innerLayout->addSpacing(10);
	innerLayout->addWidget(goButton);
	innerLayout->addStretch();
	innerLayout->addWidget(options);

	setLayout(outerLayout);
	outerLayout->addLayout(innerLayout);

	connect(addButton,SIGNAL(clicked()),this,SLOT(addPressed()));
	connect(addDirButton,SIGNAL(clicked()),this,SLOT(addDirPressed()));
	connect(removeButton,SIGNAL(clicked()),this,SLOT(removePressed()));
	connect(goButton,SIGNAL(clicked()),this,SLOT(emitGoPressed()));
}

void FileSelector::addPressed() {
	QStringList newChosen = QFileDialog::getOpenFileNames(
		this,
		tr("Select one or more files to process"),
		".",
		tr("Openable files (*.jpg *.jpeg *.jpe *.zip *.rar *.cbz *.cbr)"));
	for(int i = 0; i < newChosen.count(); ++i) {
		newChosen.replace(i,QDir::toNativeSeparators(newChosen.at(i)));
	}
	QStringList oldChosen = theChosenModel->stringList();
	oldChosen << newChosen;
	theChosenModel->setStringList(QList<QString>::fromSet(oldChosen.toSet()));
}
void FileSelector::addDirPressed() {
	QString newChosen = QFileDialog::getExistingDirectory(
		this,
		tr("Select one or more files to process"),
		".");
	newChosen = QDir::toNativeSeparators(newChosen);
	QStringList oldChosen = theChosenModel->stringList();
	oldChosen << newChosen;
	//TODO: Remove duplicates less lazily...?
	theChosenModel->setStringList(QList<QString>::fromSet(oldChosen.toSet()));
}
void FileSelector::removePressed() {
	QModelIndexList selected = theChosenOnes->selectionModel()->selectedIndexes();
	QStringList afterRemoval = theChosenModel->stringList();
	for(int i = 0; i < selected.size(); ++i) {
		afterRemoval.removeAll(theChosenModel->data(selected.at(i),Qt::DisplayRole).toString());
	}
	theChosenModel->setStringList(afterRemoval);
}
void FileSelector::emitGoPressed() {
	emit goPressed(theChosenModel->stringList(),options);
}
void FileSelector::clear() {
	theChosenModel->setStringList(QStringList());
}
