#include "MainWindow.hpp"

MainWindow::MainWindow() {
	resize(600,400);

	selector = new FileSelector(tr("&Process"),new PassedOptions);
	setCentralWidget(selector);
	connect(selector,SIGNAL(goPressed(const QStringList &,PassedOptions *)),this,
		SLOT(process(const QStringList &,PassedOptions *)));

	setWindowTitle(tr("FANG - Finally, an nbm GUI!"));
	processor = 0;
}
void MainWindow::process(const QStringList &toConvert, PassedOptions *opts) {
	if(toConvert.isEmpty()) {
		QMessageBox::warning(
			this,
			tr("FANG - Finally, an nbm GUI!"),
			tr("You have not set any files or directories to process!"),
			QMessageBox::Ok);
		return;
	}
	if(processor != 0) {
		QMessageBox::warning(
			this,
			tr("FANG - Finally, an nbm GUI!"),
			tr("Processing still hasn't exited. Please try again in a few seconds."),
			QMessageBox::Ok);
		return;
	}
	processor = new ProcessCore(toConvert,opts,this);
	QProgressDialog *procProg = new QProgressDialog(
		tr("Processing..."),
		tr("Cancel"),0,processor->count(),this);
	procProg->setWindowModality(Qt::WindowModal);
	procProg->setWindowTitle(tr("FANG - Processing..."));
	procProg->setAttribute(Qt::WA_DeleteOnClose);;
	procProg->setAttribute(Qt::WA_Resized);
	connect(processor,SIGNAL(fileDone(int)),procProg,SLOT(setValue(int)));
	connect(procProg,SIGNAL(canceled()),processor,SLOT(ceaseProcessing()));
	connect(procProg,SIGNAL(canceled()),this,SLOT(deleteProcessor()));
	connect(processor,SIGNAL(countChanged(int)),procProg,SLOT(setMaximum(int)));
	connect(processor,SIGNAL(processingDone(int,int)),procProg,SLOT(cancel()));
	connect(processor,SIGNAL(processingDone(int,int)),this,SLOT(processExit(int,int)));
	procProg->show();
	processor->beginProcessing();
}

void MainWindow::processExit(int exitType, int errors) {
	if(exitType == ProcessCore::Normal) {
		selector->clear();
		if(!errors) {
			QMessageBox::information(
				this,
				tr("FANG - Finally, an nbm GUI!"),
				tr("Files successfully converted!"));
		} else {
			QMessageBox::warning(
				this,
				tr("FANG - Finally, an nbm GUI!"),
				tr("Some files were converted, but ")+QString::number(errors)+tr(" were encountered."));
		}
	}
	else if(exitType == ProcessCore::Failed)
		QMessageBox::critical(
			this,
			tr("FANG - Finally, an nbm GUI!"),
			tr("An error occurred!"));
	deleteProcessor();
}
void MainWindow::deleteProcessor() {
	if(processor != 0) {
		delete processor;
		processor = 0;
	}
}
