#include "ProcessCore.hpp"

ProcessCore::ProcessCore(const QStringList &toFiles, PassedOptions *opts, QObject *parent)
: QObject(parent) {
	cachedCount = -1;
	currentFile = 0;
	files = toFiles;
	pct = new ProcessCoreThread(toFiles,opts,this);
	connect(pct,SIGNAL(fileDone()),this,SLOT(incrementFiles()));
	connect(pct,SIGNAL(allDone(int,int)),this,SIGNAL(processingDone(int,int)));
}
ProcessCore::~ProcessCore() {
	ceaseProcessing();
}
int ProcessCore::regenerateCount() {
	QRegExp opOnRegex("\\.(jp(eg|e|g)|cb[zr]|zip|rar)$",Qt::CaseInsensitive);
	cachedCount = 0;
	for(int i = 0; i < files.count(); ++i) {
		QString curFile = files.at(i);
		QDir d(curFile);
		if(d.isReadable()) {
			QRegExp noFollowRegex("(\\.|\\.\\.|output|temp)",Qt::CaseSensitive);
			cachedCount += getFileCount(d,opOnRegex,noFollowRegex);
		}
		if(QFile::exists(curFile) && (opOnRegex.indexIn(curFile) >= 0)) {
			++cachedCount;
		}
	}
	return cachedCount;
}
int ProcessCore::count() {
	if(cachedCount < 0) return regenerateCount();
	return cachedCount;
}
void ProcessCore::beginProcessing() {
	if(!pct->isRunning() && !pct->isFinished()) pct->start();
}
void ProcessCore::ceaseProcessing() {
	if(pct->isRunning()) {
		pct->finishUp();
		pct->wait();
	}
}
void ProcessCore::incrementFiles() {
	emit fileDone(++currentFile);
}
int ProcessCore::getFileCount(const QDir &dir, QRegExp &opOn, QRegExp &noFollow) {
	QDirIterator di(dir,QDirIterator::FollowSymlinks);
	int count = 0;
	while(di.hasNext()) {
		QString filename = di.next();
		QDir d(filename);
		if((noFollow.indexIn(filename) < 0) && d.isReadable()) {
			count += getFileCount(d,opOn,noFollow);
		}
		if(opOn.indexIn(filename) >= 0) {
			++count;
		}
	}
	return count;
}
