#include "ProcessCoreThread.hpp"
#include <QDebug>

ProcessCoreThread::ProcessCoreThread(const QStringList &toFiles, PassedOptions *opts,
		QObject *parent) : QThread(parent) {
	files = toFiles;
	finishing = false;
	options = opts;
}
void ProcessCoreThread::run() {
	unsigned errors = 0;
	if(finishing) return;
	QProcess backend;
	QRegExp jProcess("^Converting ");
	QRegExp aProcess("^Processing (zip|rar) file ");
	QRegExp aProcessDone("^Done with (zip|rar) file!");
	QRegExp errorRegex("(^Couldn't|^Error|^Out of memory|error!$)");
	QString nbm_loc;
	QString launchdir = QCoreApplication::applicationDirPath();
	#ifdef Q_WS_WIN
	if(QFile::exists(launchdir + "/nbm.exe")) nbm_loc = launchdir + "/nbm.exe";
	else nbm_loc = "nbm.exe";
	#else
	if(QFile::exists(launchdir + "/nbm")) nbm_loc = launchdir + "/nbm";
	else nbm_loc = "nbm";
	#endif

	QStringList params = options->toStringList();
	params += files;
	qDebug() << "Starting nbm with as: " << nbm_loc << params;
	backend.start(nbm_loc,params);
	if(!backend.waitForStarted(-1)) {
		emit allDone(ProcessCore::Failed);
		return;
	}
	bool arcProcessing = false;
	while(!finishing) {
		if(backend.canReadLine()) {
			QString line = backend.readLine();
			qDebug() << "nbm: " << line;
			if(errorRegex.indexIn(line) >= 0) ++errors;
			if(!arcProcessing) {
				if(jProcess.indexIn(line) >= 0) emit fileDone();
				else if(aProcess.indexIn(line) >= 0) arcProcessing = true;
			} else {
				if(aProcessDone.indexIn(line) >= 0) {
					arcProcessing = false;
					emit fileDone();
				}
			}
		}
		if(backend.waitForFinished(20)) {
			finishing = true;
			while(backend.canReadLine()) {
				QString line = backend.readLine();
				qDebug() << "nbm: " << line;
				if(errorRegex.indexIn(line) >= 0) ++errors;
			}
			emit allDone(ProcessCore::Normal,errors);
			return;
		}
	}
	emit allDone(ProcessCore::Terminated);
	backend.terminate();
}
void ProcessCoreThread::finishUp() {
	finishing = true;
}
